####################################################################
#                                                                  #
# Wilfred                                                          #
# Copyright (C) 2020, Vilhelm Prytz, <vilhelm@prytznet.se>, et al. #
#                                                                  #
# Licensed under the terms of the MIT license, see LICENSE.        #
# https://github.com/wilfred-dev/wilfred                           #
#                                                                  #
####################################################################

import threading


class KeyboardThread(threading.Thread):
    def __init__(self, input_callback, params):
        self._running = True
        self.input_callback = input_callback
        self.params = params

        super(KeyboardThread, self).__init__(name="wilfred-console-input", daemon=True)

        self.start()

    def run(self):
        while self._running:
            try:
                self.input_callback(input(), self.params)
            except Exception:
                self._running = False



####################################################################
#                                                                  #
# Wilfred                                                          #
# Copyright (C) 2020, Vilhelm Prytz, <vilhelm@prytznet.se>, et al. #
#                                                                  #
# Licensed under the terms of the MIT license, see LICENSE.        #
# https://github.com/wilfred-dev/wilfred                           #
#                                                                  #
####################################################################

import sqlite3
import click

from appdirs import user_data_dir
from os.path import isfile
from os import remove

from wilfred.message_handler import error
from wilfred.database import session, Server, EnvironmentVariable


class Migrate:
    def __init__(self):
        self._legacy_sqlite_path = f"{user_data_dir()}/wilfred/wilfred.db"

        # perform checks
        self._legacy_sqlite_db_check()

    def _legacy_sqlite_query(self, query):
        result = []

        try:
            conn = sqlite3.connect(self._legacy_sqlite_path)
            conn.row_factory = sqlite3.Row

            cur = conn.cursor()
            cur.execute(query)

            for r in cur.fetchall():
                result.append(dict(r))

            conn.commit()
            conn.close()
        except sqlite3.OperationalError as e:
            error(
                "could not communicate with database " + click.style(str(e), bold=True),
                exit_code=1,
            )
        except sqlite3.IntegrityError as e:
            error("invalid input " + click.style(str(e), bold=True), exit_code=1)

        return result

    def _legacy_sqlite_db_check(self):
        if isfile(self._legacy_sqlite_path):
            for server in self._legacy_sqlite_query("SELECT * FROM servers"):
                session.add(
                    Server(
                        id=server["id"],
                        name=server["name"],
                        image_uid=server["image_uid"],
                        memory=server["memory"],
                        port=server["port"],
                        custom_startup=server["custom_startup"],
                        status=server["status"],
                    )
                )

            try:
                session.commit()
            except Exception as e:
                error(str(e), exit_code=1)

            for variable in self._legacy_sqlite_query("SELECT * FROM variables"):
                session.add(
                    EnvironmentVariable(
                        server_id=variable["server_id"],
                        variable=variable["variable"],
                        value=variable["value"],
                    )
                )

            try:
                session.commit()
            except Exception as e:
                error(str(e), exit_code=1)

            try:
                remove(f"{self._legacy_sqlite_path}")
            except Exception as e:
                error(str(e), exit_code=1)



####################################################################
#                                                                  #
# Wilfred                                                          #
# Copyright (C) 2020, Vilhelm Prytz, <vilhelm@prytznet.se>, et al. #
#                                                                  #
# Licensed under the terms of the MIT license, see LICENSE.        #
# https://github.com/wilfred-dev/wilfred                           #
#                                                                  #
####################################################################

from pathlib import Path

from wilfred.api.images import Images
from wilfred.api.servers import Servers
from wilfred.docker_conn import docker_client
from wilfred.api.server_config import ServerConfig
from wilfred.database import Server, EnvironmentVariable, session
from wilfred.api.config_parser import Config

config = Config()
config.write(f"{str(Path.home())}/wilfred-data/servers")
config.read()

images = Images()

if not images.check_if_present():
    images.download()

images.read_images()

servers = Servers(docker_client(), config.configuration, images)


def test_create_server():
    # create
    server = Server(
        id="test",
        name="test",
        image_uid="minecraft-paper",
        memory="1024",
        port="25565",
        custom_startup=None,
        status="installing",
    )
    session.add(server)
    session.commit()

    minecraft_version = EnvironmentVariable(
        server_id=server.id, variable="MINECRAFT_VERSION", value="latest"
    )

    eula_acceptance = EnvironmentVariable(
        server_id=server.id, variable="EULA_ACCEPTANCE", value="true"
    )

    session.add(minecraft_version)
    session.add(eula_acceptance)
    session.commit()

    servers.install(server, skip_wait=False)
    servers.sync()


def test_start_server():
    server = session.query(Server).filter_by(id="test").first()

    if server.status == "installing":
        raise Exception("server is installing")

    servers.set_status(server, "running")
    servers.sync()


def test_pseudo_config_write():
    server = session.query(Server).filter_by(id="test").first()

    image = images.get_image(server.image_uid)

    Path(f"{str(Path.home())}/temp/test_test").mkdir(parents=True, exist_ok=True)

    with open(f"{str(Path.home())}/temp/test_test/server.properties", "w") as f:
        f.write("\n".join(("query.port=25564", "server-port=25564")))

    ServerConfig(
        {"data_path": f"{str(Path.home())}/temp"}, servers, server, image
    ).write_environment_variables()


def test_delete_server():
    server = session.query(Server).filter_by(id="test").first()
    servers.remove(server)



####################################################################
#                                                                  #
# Wilfred                                                          #
# Copyright (C) 2020, Vilhelm Prytz, <vilhelm@prytznet.se>, et al. #
#                                                                  #
# Licensed under the terms of the MIT license, see LICENSE.        #
# https://github.com/wilfred-dev/wilfred                           #
#                                                                  #
####################################################################

import requests
import click

from random import choice
from string import ascii_lowercase, digits

from wilfred.version import version, commit_hash
from wilfred.message_handler import warning


def random_string(length=8):
    """
    Generate a random string of fixed length

    :param int length: length of string to generate
    """

    return "".join(choice(ascii_lowercase + digits) for i in range(length))


def check_for_new_releases(enable_emojis=True):
    """
    Checks if a new version is available on GitHub
    """

    url = "https://api.github.com/repos/wilfred-dev/wilfred/tags"
    key = "name"
    version_type = "version"

    if version == "0.0.0.dev0":
        url = "https://api.github.com/repos/wilfred-dev/wilfred/commits"
        key = "sha"
        version_type = "commit"

    r = requests.get(url)

    if r.status_code != requests.codes.ok:
        warning("unable to retrieve latest version")

        return

    try:
        latest = r.json()[0][key]
    except Exception:
        warning("unable to parse release data")

        return

    compare = commit_hash if version == "0.0.0.dev0" else f"v{version}"

    if latest != compare:
        click.echo(
            "".join(
                (
                    f"{'🎉 ' if enable_emojis else click.style('! ', fg='green')}",
                    f"A new {version_type} of Wilfred is available! {latest}",
                )
            )
        )

    return


def is_integer(variable):
    try:
        int(variable)
    except Exception:
        return False

    return True


def set_in_dict(dic, keys, value):
    for key in keys[:-1]:
        dic = dic.setdefault(key, {})
    dic[keys[-1]] = value

    return dic



####################################################################
#                                                                  #
# Wilfred                                                          #
# Copyright (C) 2020, Vilhelm Prytz, <vilhelm@prytznet.se>, et al. #
#                                                                  #
# Licensed under the terms of the MIT license, see LICENSE.        #
# https://github.com/wilfred-dev/wilfred                           #
#                                                                  #
####################################################################

from functools import wraps

from wilfred.message_handler import error
from wilfred.api.config_parser import Config, NoConfiguration

config = Config()

try:
    config.read()
except NoConfiguration:
    pass


def configuration_present(f):
    @wraps(f)
    def decorated_function(*args, **kwargs):
        if not config.configuration:
            error("Wilfred has not been configured", exit_code=1)
        return f(*args, **kwargs)

    return decorated_function



####################################################################
#                                                                  #
# Wilfred                                                          #
# Copyright (C) 2020, Vilhelm Prytz, <vilhelm@prytznet.se>, et al. #
#                                                                  #
# Licensed under the terms of the MIT license, see LICENSE.        #
# https://github.com/wilfred-dev/wilfred                           #
#                                                                  #
####################################################################

# use "0.0.0.dev0" to indicate incremental commit build
# specify version "0.3.0" to indicate standard release
# "commit_date" should be replaced upon building
version = "0.8.0"
commit_hash = "22fb139fc1538ad229ceeaba3a57f3f174b62d16"
commit_date = "2020-12-19"



####################################################################
#                                                                  #
# Wilfred                                                          #
# Copyright (C) 2020, Vilhelm Prytz, <vilhelm@prytznet.se>, et al. #
#                                                                  #
# Licensed under the terms of the MIT license, see LICENSE.        #
# https://github.com/wilfred-dev/wilfred                           #
#                                                                  #
####################################################################

import click

from tabulate import tabulate

from wilfred.container_variables import ContainerVariables
from wilfred.errors import WilfredException, ParseError, WriteError

from wilfred.api.parser.properties import properties_read, properties_write
from wilfred.api.parser.yaml import yaml_read, yaml_write
from wilfred.api.parser.json import json_read, json_write


class UnsupportedFiletype(WilfredException):
    """File type is not supported by parser"""


class ServerConfig:
    def __init__(self, configuration, servers, server, image):
        """
        Read, edit and update configurations for servers. Exposes server specific configs to Wilfred.

        :param configuration: JSON dict for Wilfred config
        :param object servers: wilfred.api.servers object
        :param str server: SQLAlchemy server object
        :param dictionary image: Image dict object
        """

        self._configuration = configuration
        self._servers = servers
        self._server = server
        self._image = image

        self.raw = []
        self._variables = []  # list of dicts

        self._parse()

    def _parse(self):
        """iterates configuration files for the specific server and parses the files"""

        def _err(e):
            raise ParseError(f"failed to parse {file['filename']}, err {str(e)}")

        for file in self._image["config"]["files"]:
            path = f"{self._configuration['data_path']}/{self._server.name}_{self._server.id}/{file['filename']}"

            if file["parser"] == "properties":
                try:
                    _raw = properties_read(path)
                except Exception as e:
                    _err(e)

                _raw["_wilfred_config_filename"] = file["filename"]
                self.raw.append(_raw)

                continue

            if file["parser"] == "yaml":
                try:
                    _raw = yaml_read(path)
                except Exception as e:
                    _err(e)

                _raw["_wilfred_config_filename"] = file["filename"]
                self.raw.append(_raw)

                continue

            if file["parser"] == "json":
                try:
                    _raw = json_read(path)
                except Exception as e:
                    _err(e)

                _raw["_wilfred_config_filename"] = file["filename"]
                self.raw.append(_raw)

                continue

            raise UnsupportedFiletype(file["parser"])
        return True

    def pretty(self):
        """returns parsed configuration variables in a print-friendly format"""

        headers = {
            "file": click.style("Config File", bold=True),
            "setting": click.style("Setting", bold=True),
            "value": click.style("Value", bold=True),
        }

        data = []

        for file in self.raw:
            for k, v in file.items():
                for _image_config_file in self._image["config"]["files"]:
                    for x in _image_config_file["environment"]:
                        k = (
                            f"{k} ({click.style('not editable', fg='red', bold=True)})"
                            if x["config_variable"] == k
                            else k
                        )

                data.append(
                    {"file": file["_wilfred_config_filename"], "setting": k, "value": v}
                ) if k != "_wilfred_config_filename" else None

        return tabulate(
            data,
            headers=headers,
            tablefmt="fancy_grid",
        )

    def edit(self, filename, variable, value, override_linking_check=False):
        """modifies value of specified variable"""

        try:
            value = int(value)
        except Exception:
            pass

        try:
            value = True if value.lower() == "true" else value
            value = False if value.lower() == "false" else value
        except Exception:
            pass

        for file in self._image["config"]["files"]:
            if file["filename"] == filename:
                path = f"{self._configuration['data_path']}/{self._server.name}_{self._server.id}"

                for _image_config_file in self._image["config"]["files"]:
                    for x in _image_config_file["environment"]:
                        if (
                            x["config_variable"] == variable
                            and not override_linking_check
                        ):
                            raise WriteError(
                                "This setting is linked to an environment variable and is therefore not editable directly"
                            )

                if file["parser"] == "properties":
                    properties_write(f"{path}/{file['filename']}", variable, value)

                if file["parser"] == "yaml":
                    yaml_write(f"{path}/{file['filename']}", variable, value)

                if file["parser"] == "json":
                    json_write(f"{path}/{file['filename']}", variable, value)

                if variable in file["action"]:
                    self._servers.command(
                        self._server, file["action"][variable].format(value)
                    )

    def write_environment_variables(self):
        """writes environment variable to config file(s)"""

        for file in self._image["config"]["files"]:
            for _env in file["environment"]:
                env_vars = ContainerVariables(self._server, self._image).get_env_vars()

                if _env["environment_variable"] in env_vars:
                    self.edit(
                        file["filename"],
                        _env["config_variable"],
                        _env["value_format"].format(
                            env_vars[_env["environment_variable"]]
                        )
                        if _env["value_format"]
                        else env_vars[_env["environment_variable"]],
                        override_linking_check=True,
                    )



####################################################################
#                                                                  #
# Wilfred                                                          #
# Copyright (C) 2020, Vilhelm Prytz, <vilhelm@prytznet.se>, et al. #
#                                                                  #
# Licensed under the terms of the MIT license, see LICENSE.        #
# https://github.com/wilfred-dev/wilfred                           #
#                                                                  #
####################################################################

from configparser import RawConfigParser


def _config_parser(path):
    """appends a dummy section to the top of the raw file so that the built-in
    python function, ConfigParser, can read and parse the file"""

    with open(path) as f:
        file_content = "[dummy_section]\n" + f.read()

    config_parser = RawConfigParser()
    config_parser.read_string(file_content)

    return config_parser


def properties_read(path):
    config_parser = _config_parser(path)

    settings = {}

    for item in config_parser.items("dummy_section"):
        settings[item[0]] = item[1]

    return settings


def properties_write(path, key, value):
    with open(path) as f:
        raw = f.read().split("\n")

    _write = []

    for line in raw:
        if key in line:
            if line.split("=")[0] == key:
                _write.append(f"{key}={value}")
                continue

        _write.append(line)

    with open(path, "w") as f:
        f.write("\n".join(_write))



####################################################################
#                                                                  #
# Wilfred                                                          #
# Copyright (C) 2020, Vilhelm Prytz, <vilhelm@prytznet.se>, et al. #
#                                                                  #
# Licensed under the terms of the MIT license, see LICENSE.        #
# https://github.com/wilfred-dev/wilfred                           #
#                                                                  #
####################################################################

import docker


def docker_client(base_url=None):
    """returns client object used for Docker socket communication

    :param str base_url: Base URL for Docker socket. Uses env if None."""

    client = docker.DockerClient(base_url=base_url) if base_url else docker.from_env()

    return client



####################################################################
#                                                                  #
# Wilfred                                                          #
# Copyright (C) 2020, Vilhelm Prytz, <vilhelm@prytznet.se>, et al. #
#                                                                  #
# Licensed under the terms of the MIT license, see LICENSE.        #
# https://github.com/wilfred-dev/wilfred                           #
#                                                                  #
####################################################################

from sqlalchemy import create_engine, Column, Integer, String, ForeignKey
from sqlalchemy.orm import relationship, sessionmaker, validates
from sqlalchemy.ext.declarative import declarative_base

from appdirs import user_data_dir
from os.path import isdir
from pathlib import Path


if not isdir(f"{user_data_dir()}/wilfred"):
    Path(f"{user_data_dir()}/wilfred").mkdir(parents=True, exist_ok=True)

database_path = f"{user_data_dir()}/wilfred/wilfred.sqlite"
engine = create_engine(f"sqlite:///{database_path}")
Base = declarative_base()


class Server(Base):
    __tablename__ = "servers"

    id = Column(String, primary_key=True, unique=True)
    name = Column(String(20), unique=True)
    image_uid = Column(String)
    memory = Column(Integer)
    port = Column(Integer, unique=True)
    custom_startup = Column(String)
    status = Column(String)

    environment_variables = relationship("EnvironmentVariable")

    # SQLite does not enforce string maximums
    @validates("name")
    def validate_name(self, key, name) -> str:
        if len(name) > 20:
            raise ValueError("name is too long (max 20 characters)")
        return name

    def __repr__(self):
        return f"<Server(id='{self.id}', name='{self.name}', image_uid='{self.image_uid}', port='{self.port}')>"


class EnvironmentVariable(Base):
    __tablename__ = "environment_variables"

    id = Column(Integer, primary_key=True)
    server_id = Column(String, ForeignKey("servers.id"), unique=False)
    variable = Column(String)
    value = Column(String)


Base.metadata.create_all(engine)

Session = sessionmaker()
Session.configure(bind=engine)
session = Session()



####################################################################
#                                                                  #
# Wilfred                                                          #
# Copyright (C) 2020, Vilhelm Prytz, <vilhelm@prytznet.se>, et al. #
#                                                                  #
# Licensed under the terms of the MIT license, see LICENSE.        #
# https://github.com/wilfred-dev/wilfred                           #
#                                                                  #
####################################################################

import json

from appdirs import user_config_dir
from os.path import isfile, isdir
from pathlib import Path

from wilfred.errors import WilfredException, ParseError, WriteError

API_VERSION = 1


class NoConfiguration(WilfredException):
    """Configuration files does not exist"""


class ConfigurationAPIMismatch(WilfredException):
    """API level of config does not match with API level present in config"""


class Config(object):
    def __init__(self):
        self.data_dir = f"{user_config_dir()}/wilfred"
        self.config_path = f"{self.data_dir}/config.json"
        self.configuration = None

        if not isdir(self.data_dir):
            Path(self.data_dir).mkdir(parents=True, exist_ok=True)

    def read(self):
        if not isfile(self.config_path):
            raise NoConfiguration

        with open(self.config_path) as f:
            try:
                self.configuration = json.loads(f.read())
            except Exception as e:
                raise ParseError(e)

        if self.configuration["meta"]["version"] != API_VERSION:
            raise ConfigurationAPIMismatch(
                f"Wilfred has version {API_VERSION}, file has version {self.configuration['meta']['version']}"
            )

    def write(self, data_path):
        try:
            Path(data_path).mkdir(parents=True, exist_ok=True)
        except Exception as e:
            raise WriteError(e)

        with open(self.config_path, "w") as f:
            f.write(
                json.dumps(
                    {"meta": {"version": API_VERSION}, "data_path": data_path}, indent=4
                )
            )



####################################################################
#                                                                  #
# Wilfred                                                          #
# Copyright (C) 2020, Vilhelm Prytz, <vilhelm@prytznet.se>, et al. #
#                                                                  #
# Licensed under the terms of the MIT license, see LICENSE.        #
# https://github.com/wilfred-dev/wilfred                           #
#                                                                  #
####################################################################


class WilfredException(Exception):
    """
    A base class from which all other exceptions inherit.

    If you want to catch all errors that the Wilfred API might raise,
    catch this base exception. Though, dependencies of Wilfred might
    raise other exceptions which this base exception does not cover.
    """


class ReadError(WilfredException):
    """
    Exception occured while reading file
    """


class ParseError(WilfredException):
    """
    Exception occured while parsing file (malformed or missing variables)
    """


class WriteError(WilfredException):
    """
    Exception occured while wiring to file (permission denied or invalid path)
    """



####################################################################
#                                                                  #
# Wilfred                                                          #
# Copyright (C) 2020, Vilhelm Prytz, <vilhelm@prytznet.se>, et al. #
#                                                                  #
# Licensed under the terms of the MIT license, see LICENSE.        #
# https://github.com/wilfred-dev/wilfred                           #
#                                                                  #
####################################################################

import yaml

# from wilfred.core import is_integer


def yaml_read(path):  # this function should be refactored later on!!!
    with open(path) as f:
        _raw = yaml.load(f.read(), Loader=yaml.FullLoader)

    _reformatted = {}

    def _iterate_dict(data, _name):
        """
        iterates a dictionary and continues to iterate the value of the dict if the
        type is dictionary, list or tuple (recursion). Adds the dictionary key and value
        to the _reformatted dict if it's a string, integer or boolean
        """

        _def = _name
        for k, v in data.items():
            _name = f"{_def}/{k}"
            if type(v) in [dict]:
                _iterate_dict(v, _name)
            if type(v) in [list, tuple]:
                _iterate_list(v, _name)

            if type(v) in [str, int, bool]:
                _reformatted[f"{_def}/{k}"] = v

    def _iterate_list(data, _name):
        """
        iterates a list and continues to iterate the values of the list if the
        type is dictionary, list or tuple (recursion). Adds the list value
        to the _reformatted dict if it's a string, integer or boolean
        """

        _def = _name

        i = 0
        for x in data:
            _name = f"{_def}/{i}"
            if type(x) in [dict]:
                _iterate_dict(x, _name)
            if type(x) in [list, tuple]:
                _iterate_list(x, _name)

            if type(x) in [str, int, bool]:
                _reformatted[f"{_def}/{i}"] = x

            i = i + 1

    for k, v in _raw.items():
        _name = k

        if type(v) in [dict]:
            _iterate_dict(v, _name)
        if type(v) in [list, tuple]:
            _iterate_list(v, _name)

        if type(v) in [str, int, bool]:
            _reformatted[k] = v

    return _reformatted


# def yaml_write(path, key, value): # does not work yet!!
#     with open(path) as f:
#         _raw = yaml.load(f.read(), Loader=yaml.FullLoader)

#     def _traverse(obj, path=None, callback=None):
#         if path is None:
#             path = []

#         if type(obj) in [dict]:
#             value = {k: _traverse(v, path + [k], callback) for k, v in obj.items()}
#         elif type(obj) in [list, tuple]:
#             value = [_traverse(elem, path + [[]], callback) for elem in obj]
#         else:
#             value = obj

#         if callback is None:
#             return value
#         else:
#             return callback(path, value)

#     def _traverse_modify(obj, target_path, action):
#         def transformer(path, value):
#             if path == target_path:
#                 return action(value)
#             else:
#                 return value

#         return _traverse(obj, callback=transformer)

#     def _callback(v):
#         return value

#     _reformatted_path = [[] if is_integer(x) else x for x in key.split("/")]
#     _traverse_modify(_raw, _reformatted_path, _callback)

#     # with open(path, "w") as f:
#         # yaml.dump(_raw, f)


def yaml_write(path, key, value):  # does not work yet!!
    raise Exception("Modifying YAML variables is currently not supported")

    # with open(path) as f:
    #     _raw = yaml.load(f.read(), Loader=yaml.FullLoader)

    # def _entrypoint(search, current, value, parent):
    #     if type(value) in [dict]:
    #         current = current+1
    #         parent = value
    #         _iterate_dict(search, current, value, parent)
    #     if type(value) in [list, tuple]:
    #         current = current+1
    #         parent = value
    #         _iterate_list(search, current, value, parent)
    #     if type(value) in [str, int, bool]:
    #         print(f"{search[current]} with value {value} in {parent}")
    #         exit(1)

    # def _iterate_dict(search, current, value, parent):
    #     for k, v in value.items():
    #         if k == search[current]:
    #             _entrypoint(search, current, v, parent)

    # def _iterate_list(search, current, value, parent):
    #     _entrypoint(search, current, value[search[current]], parent)

    # _search = [int(x) if is_integer(x) else x for x in key.split("/")]
    # _current = 0

    # for i in range(len(_search)):
    #     for k, v in _raw.items():
    #         if _search[_current] == k:
    #             _entrypoint(_search, _current, v, _raw)

    # with open(path, "w") as f:
    #     yaml.dump(_raw)

    # return True



####################################################################
#                                                                  #
# Wilfred                                                          #
# Copyright (C) 2020, Vilhelm Prytz, <vilhelm@prytznet.se>, et al. #
#                                                                  #
# Licensed under the terms of the MIT license, see LICENSE.        #
# https://github.com/wilfred-dev/wilfred                           #
#                                                                  #
####################################################################

import json

from appdirs import user_config_dir
from pathlib import Path
from os.path import isdir, join, isfile
from os import walk, remove
from requests import get
from zipfile import ZipFile
from shutil import move, rmtree
from copy import deepcopy
from datetime import datetime, timedelta

from wilfred.errors import WilfredException, ReadError, ParseError
from wilfred.version import version

API_VERSION = 2


class ImagesNotPresent(WilfredException):
    """Default images not present on host"""


class ImagesNotRead(WilfredException):
    """Images are not read yet"""


class ImageAPIMismatch(WilfredException):
    """API level of image and API level of Wilfred mismatch"""


class ImagesOutdated(WilfredException):
    """Wilfred images are outdated and require refresh"""


class Images(object):
    """Manage Wilfred images"""

    def __init__(self):
        self.config_dir = f"{user_config_dir()}/wilfred"
        self.image_dir = f"{self.config_dir}/images"
        self.images = []

        if not isdir(self.image_dir):
            Path(self.image_dir).mkdir(parents=True, exist_ok=True)

    def download(self, branch="master", repo="wilfred-dev/images"):
        """
        Downloads default Wilfred Images from GitHub
        """

        rmtree(f"{self.image_dir}/default", ignore_errors=True)

        with open(f"{self.config_dir}/img.zip", "wb") as f:
            response = get(
                f"https://github.com/{repo}/archive/{branch}.zip",
                stream=True,
            )
            f.write(response.content)

        with ZipFile(f"{self.config_dir}/img.zip", "r") as obj:
            obj.extractall(f"{self.config_dir}/temp_images")

        move(
            f"{self.config_dir}/temp_images/images-{branch}/images",
            f"{self.image_dir}/default",
        )

        remove(f"{self.config_dir}/img.zip")
        rmtree(f"{self.config_dir}/temp_images")

        # write to cache info that images have been updated
        data = {"time": str(datetime.now()), "version": version}

        with open(f"{self.config_dir}/image_cache.json", "w") as f:
            json.dump(data, f)

    def data_strip_non_ui(self):
        """
        Returns a list of all images with only the variables important to the user shown

        Returns:
            Returns ``list`` of images.

        Raises:
            :py:class:`wilfred.api.images.ImagesNotRead`
        """

        if not self._check_if_read():
            raise ImagesNotRead("Read images before trying to get images")

        _images = deepcopy(self.images)

        for d in _images:
            for key in (
                "meta",
                "installation",
                "docker_image",
                "command",
                "stop_command",
                "variables",
                "user",
                "config",
            ):
                try:
                    del d[key]
                except Exception:
                    pass

        return _images

    def get_image(self, uid: str):
        """
        Retrieves image configuration for specific image

        Returns:
            Returns ``dict`` of image configuration.

        Raises:
            :py:class:`wilfred.api.images.ImagesNotRead`
        """

        if not self._check_if_read():
            raise ImagesNotRead("Read images before trying to get image")

        return next(filter(lambda img: img["uid"] == uid, self.images), None)

    def read_images(self):
        """
        Reads and parses all images on system

        Returns:
            Returns ``True`` if success.

        Raises:
            :py:class:`wilfred.api.images.ImagesNotPresent`
            :py:class:`wilfred.api.errors.ReadError`
            :py:class:`wilfred.api.images.ImageAPIMismatch`
        """

        if not self.check_if_present():
            raise ImagesNotPresent("Default images not present")

        if self.is_outdated():
            raise ImagesOutdated("Images are outdated, refresh required")

        self.image_fetch_date = "N/A"
        self.image_fetch_version = "N/A"
        self.image_time_to_refresh = "N/A"

        try:
            with open(f"{self.config_dir}/image_cache.json") as f:
                data = json.load(f)

                self.image_fetch_date = datetime.strptime(
                    data["time"], "%Y-%m-%d %H:%M:%S.%f"
                )
                self.image_fetch_version = data["version"]
                self.image_time_to_refresh = timedelta(days=7) - (
                    datetime.now() - self.image_fetch_date
                )
        except Exception:
            pass

        self.images = []

        for root, dirs, files in walk(self.image_dir):
            for file in files:
                if file.endswith(".json"):
                    with open(join(root, file)) as f:
                        try:
                            _image = json.loads(f.read())
                        except Exception as e:
                            raise ReadError(f"{file} failed with exception {str(e)}")

                        try:
                            if _image["meta"]["api_version"] != API_VERSION:
                                raise ImageAPIMismatch(
                                    " ".join(
                                        (
                                            f"{file} API level {_image['meta']['api_version']},",
                                            f"Wilfred API level {API_VERSION}",
                                        )
                                    )
                                )
                        except ImageAPIMismatch as e:
                            raise ImageAPIMismatch(str(e))
                        except Exception as e:
                            raise ReadError(f"{file} with err {str(e)}")

                        self._verify(_image, file)
                        self.images.append(_image)

        return True

    def check_if_present(self):
        """Checks if default images are present"""

        if not isdir(f"{self.image_dir}/default"):
            return False

        return True

    def is_outdated(self):
        """Checks if default images are outdated"""

        if not isfile(f"{self.config_dir}/image_cache.json"):
            return True

        try:
            with open(f"{self.config_dir}/image_cache.json") as f:
                data = json.load(f)

                if (
                    datetime.now()
                    - datetime.strptime(data["time"], "%Y-%m-%d %H:%M:%S.%f")
                ) > timedelta(days=7):
                    return True

                if data["version"] != version:
                    return True

                return False
        except Exception:
            return True

    def _verify(self, image: dict, file: str):
        def _exception(key):
            raise ParseError(f"image {file} is missing key {str(key)}")

        for key in (
            "meta",
            "uid",
            "name",
            "author",
            "docker_image",
            "command",
            "user",
            "stop_command",
            "default_image",
            "variables",
            "installation",
            "config",
        ):
            try:
                image[key]
            except Exception:
                return _exception(key)

        if image["uid"] != image["uid"].lower():
            raise ParseError(f"image {file} uid must be lowercase")

        for key in ["api_version"]:
            try:
                image["meta"][key]
            except Exception:
                return _exception(key)

        for key in ("docker_image", "shell", "script"):
            try:
                image["installation"][key]
            except Exception:
                return _exception(key)

        try:
            image["config"]["files"]
        except Exception:
            return _exception(key)

        if len(image["config"]["files"]) > 0:
            for i in range(len(image["config"]["files"])):
                for key in ("filename", "parser", "environment", "action"):
                    try:
                        image["config"]["files"][i][key]
                    except Exception:
                        return _exception(key)

                # check for valid syntax in environment variables
                for x in range(len(image["config"]["files"][i]["environment"])):
                    for key in (
                        "config_variable",
                        "environment_variable",
                        "value_format",
                    ):
                        try:
                            image["config"]["files"][i]["environment"][x][key]
                        except Exception:
                            return _exception(
                                f"{image['config']['files'][i]['filename']} environment key {key}"
                            )

            # should also check for valid syntax in environment variable linking

        if len(image["variables"]) > 0:
            for i in range(len(image["variables"])):
                for key in ("prompt", "variable", "install_only", "default", "hidden"):
                    try:
                        image["variables"][i][key]
                    except Exception:
                        return _exception(key)

        return True

    def _check_if_read(self):
        if len(self.images) == 0:
            return False

        return True



####################################################################
#                                                                  #
# Wilfred                                                          #
# Copyright (C) 2020, Vilhelm Prytz, <vilhelm@prytznet.se>, et al. #
#                                                                  #
# Licensed under the terms of the MIT license, see LICENSE.        #
# https://github.com/wilfred-dev/wilfred                           #
#                                                                  #
####################################################################

from click.testing import CliRunner
from wilfred.wilfred import cli


def test_basic():
    runner = CliRunner()
    result = runner.invoke(cli)

    assert result.exit_code == 0


def test_version():
    runner = CliRunner()
    result = runner.invoke(cli, "--version")

    assert result.exit_code == 0


def test_path():
    runner = CliRunner()
    result = runner.invoke(cli, "--path")

    assert result.exit_code == 0



####################################################################
#                                                                  #
# Wilfred                                                          #
# Copyright (C) 2020, Vilhelm Prytz, <vilhelm@prytznet.se>, et al. #
#                                                                  #
# Licensed under the terms of the MIT license, see LICENSE.        #
# https://github.com/wilfred-dev/wilfred                           #
#                                                                  #
####################################################################

from wilfred.database import session, EnvironmentVariable


class ContainerVariables(object):
    def __init__(self, server, image, install=False):
        self._server = server
        self._image = image
        self._install = install

    def parse_startup_command(self, cmd):
        for k, v in self.get_env_vars().items():
            cmd = cmd.replace("{{image.env." + str(k) + "}}", str(v if v else ""))

        return cmd

    def get_env_vars(self):
        environment = {}

        for var in self._image["variables"]:
            value = (
                session.query(EnvironmentVariable)
                .filter_by(server_id=self._server.id)
                .filter_by(variable=var["variable"])
                .first()
                .value
            )

            if var["install_only"] and not self._install:
                continue

            environment[var["variable"]] = value

        environment["SERVER_MEMORY"] = self._server.memory
        environment["SERVER_PORT"] = self._server.port

        return environment



####################################################################
#                                                                  #
# Wilfred                                                          #
# Copyright (C) 2020, Vilhelm Prytz, <vilhelm@prytznet.se>, et al. #
#                                                                  #
# Licensed under the terms of the MIT license, see LICENSE.        #
# https://github.com/wilfred-dev/wilfred                           #
#                                                                  #
####################################################################

import sys
from setuptools import setup, find_packages
from wilfred.version import version

assert sys.version_info >= (3, 6, 0), "Wilfred requires Python 3.6+"

with open("README.md", "r", encoding="utf-8") as f:
    long_description = f.read()

setup(
    name="wilfred",
    version=version,
    author="Vilhelm Prytz",
    author_email="vlhelm@prytznet.se",
    description="A CLI for managing game servers using Docker.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://wilfredproject.org",
    project_urls={
        "Documentation": "https://docs.wilfredproject.org",
        "Code": "https://github.com/wilfred-dev/wilfred",
        "Issue tracker": "https://github.com/wilfred-dev/wilfred/issues",
    },
    packages=find_packages(),
    license="MIT",
    classifiers=[
        "Programming Language :: Python",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3 :: Only",
        "Environment :: Console",
        "License :: OSI Approved :: MIT License",
        "Operating System :: MacOS :: MacOS X",
        "Operating System :: POSIX :: Linux",
        "Operating System :: Unix",
        "Operating System :: Microsoft :: Windows",
        "Natural Language :: English",
    ],
    python_requires=">=3.6",
    install_requires=[
        "docker",
        "click",
        "colorama",
        "appdirs",
        "requests",
        "tabulate",
        "halo",
        "sqlalchemy",
        "pyyaml",
        "pypiwin32 ; platform_system=='Windows'",
    ],
    entry_points={"console_scripts": ["wilfred=wilfred.wilfred:main"]},
)



####################################################################
#                                                                  #
# Wilfred                                                          #
# Copyright (C) 2020, Vilhelm Prytz, <vilhelm@prytznet.se>, et al. #
#                                                                  #
# Licensed under the terms of the MIT license, see LICENSE.        #
# https://github.com/wilfred-dev/wilfred                           #
#                                                                  #
####################################################################

import json


def json_read(path):
    with open(path) as f:
        _raw = json.loads(f.read())

    _reformatted = {}

    def _iterate_dict(data, _name):
        """
        iterates a dictionary and continues to iterate the value of the dict if the
        type is dictionary, list or tuple (recursion). Adds the dictionary key and value
        to the _reformatted dict if it's a string, integer or boolean
        """

        _def = _name
        for k, v in data.items():
            _name = f"{_def}/{k}"
            if type(v) in [dict]:
                _iterate_dict(v, _name)
            if type(v) in [list, tuple]:
                _iterate_list(v, _name)

            if type(v) in [str, int, bool]:
                _reformatted[f"{_def}/{k}"] = v

    def _iterate_list(data, _name):
        """
        iterates a list and continues to iterate the values of the list if the
        type is dictionary, list or tuple (recursion). Adds the list value
        to the _reformatted dict if it's a string, integer or boolean
        """

        _def = _name

        i = 0
        for x in data:
            _name = f"{_def}/{i}"
            if type(x) in [dict]:
                _iterate_dict(x, _name)
            if type(x) in [list, tuple]:
                _iterate_list(x, _name)

            if type(x) in [str, int, bool]:
                _reformatted[f"{_def}/{i}"] = x

            i = i + 1

    for k, v in _raw.items():
        _name = k

        if type(v) in [dict]:
            _iterate_dict(v, _name)
        if type(v) in [list, tuple]:
            _iterate_list(v, _name)

        if type(v) in [str, int, bool]:
            _reformatted[k] = v

    return _reformatted


def json_write(path, key, value):
    raise Exception("Modifying JSON variables is currently not supported")

    # return True



####################################################################
#                                                                  #
# Wilfred                                                          #
# Copyright (C) 2020, Vilhelm Prytz, <vilhelm@prytznet.se>, et al. #
#                                                                  #
# Licensed under the terms of the MIT license, see LICENSE.        #
# https://github.com/wilfred-dev/wilfred                           #
#                                                                  #
####################################################################

import click
import docker

from pathlib import Path
from shutil import rmtree
from os import remove as remove_file
from os import rename
from time import sleep
from sys import platform
from subprocess import call
from sqlalchemy import inspect

from wilfred.database import session, Server, EnvironmentVariable
from wilfred.keyboard import KeyboardThread
from wilfred.container_variables import ContainerVariables
from wilfred.api.images import Images
from wilfred.errors import WilfredException, WriteError


class ServerNotRunning(WilfredException):
    """Server is not running"""


class Servers(object):
    def __init__(
        self, docker_client: docker.DockerClient, configuration: dict, images: Images
    ):
        """
        Initiates wilfred.api.Servers, method for controlling servers

        Args:
            docker_client (docker.DockerClient): DockerClient object from Docker module
            configuration (dict): Dictionary of Wilfred config
            images (Images): wilfred.api.Images object
        """

        self._images = images
        self._configuration = configuration
        self._docker_client = docker_client

    def all(self, cpu_load=False, memory_usage=False):
        """
        Returns data of all servers

        Args:
            cpu_load (bool): Include the CPU load of the container. Defaults to `None` if server is not running.
            memory_usage (bool): Include RAM usage of the container. Defaults to `None` if server is not running.
        """

        servers = [
            {c.key: getattr(u, c.key) for c in inspect(u).mapper.column_attrs}
            for u in session.query(Server).all()
        ]

        for server in servers:
            if cpu_load or memory_usage:
                for server in servers:
                    _running = True

                    try:
                        container = self._docker_client.containers.get(
                            f"wilfred_{server['id']}"
                        )
                        d = container.stats(stream=False)
                    except docker.errors.NotFound:
                        server.update({"cpu_load": "-"})
                        server.update({"memory_usage": "-"})
                        _running = False
                    except Exception:
                        server.update({"cpu_load": "error"})
                        server.update({"memory_usage": "error"})
                        _running = False

                    if cpu_load and _running:
                        cpu_count = len(d["cpu_stats"]["cpu_usage"]["percpu_usage"])
                        cpu_percent = 0.0
                        cpu_delta = float(
                            d["cpu_stats"]["cpu_usage"]["total_usage"]
                        ) - float(d["precpu_stats"]["cpu_usage"]["total_usage"])
                        system_delta = float(
                            d["cpu_stats"]["system_cpu_usage"]
                        ) - float(d["precpu_stats"]["system_cpu_usage"])
                        if system_delta > 0.0:
                            cpu_percent = f"{round(cpu_delta / system_delta * 100.0 * cpu_count)}%"

                        server.update({"cpu_load": cpu_percent if cpu_percent else "-"})

                    if memory_usage and _running:
                        server.update(
                            {
                                "memory_usage": f"{round(d['memory_stats']['usage'] / 10**6)} MB"
                            }
                        )

        return servers

    def set_status(self, server, status):
        server.status = status
        session.commit()

    def sync(self):
        """
        Performs sync, checks for state of containers
        """

        for server in session.query(Server).all():
            if server.status == "installing":
                try:
                    self._docker_client.containers.get(f"wilfred_{server.id}")
                except docker.errors.NotFound:
                    self.set_status(server, "stopped")

            # stopped
            if server.status == "stopped":
                self._stop(server)

            # start
            if server.status == "running":
                try:
                    self._docker_client.containers.get(f"wilfred_{server.id}")
                except docker.errors.NotFound:
                    self._start(server)

    def remove(self, server: Server):
        """
        Removes specified server

        Args:
            server (wilfred.database.Server): Server database object
        """

        path = f"{self._configuration['data_path']}/{server.name}_{server.id}"

        for x in (
            session.query(EnvironmentVariable).filter_by(server_id=server.id).all()
        ):
            session.delete(x)

        session.delete(server)
        session.commit()

        try:
            container = self._docker_client.containers.get(f"wilfred_{server.id}")
            container.kill()
        except docker.errors.NotFound:
            pass

        rmtree(path, ignore_errors=True)

    def console(self, server: Server, disable_user_input=False):
        """
        Enters server console

        Args:
            server (wilfred.database.Server): Server database object
            disable_user_input (bool): Blocks user input if `True`. By default this is `False`.

        Raises:
            :py:class:`ServerNotRunning`
                If server is not running
        """

        try:
            container = self._docker_client.containers.get(f"wilfred_{server.id}")
        except docker.errors.NotFound:
            raise ServerNotRunning(f"server {server.id} is not running")

        if platform.startswith("win"):
            click.echo(container.logs())
            call(["docker", "attach", container.id])
        else:
            if not disable_user_input:
                KeyboardThread(self._console_input_callback, params=server)

            try:
                for line in container.logs(stream=True, tail=200):
                    click.echo(line.strip())
            except docker.errors.NotFound:
                raise ServerNotRunning(f"server {server.id} is not running")

    def install(self, server: Server, skip_wait=False, spinner=None):
        """
        Performs installation

        Args:
            server (wilfred.database.Server): Server database object
            skip_wait (bool): Doesn't stall while waiting for server installation to complete if `True`.
            spinner (Halo): If `Halo` spinner object is defined, will then write and perform actions to it.

        Raises:
            :py:class:`WriteError`
                If not able to create directory or write to it
        """

        path = f"{self._configuration['data_path']}/{server.name}_{server.id}"
        image = self._images.get_image(server.image_uid)

        if platform.startswith("win"):
            path = path.replace("/", "\\")

        try:
            Path(path).mkdir(parents=True, exist_ok=True)
        except Exception as e:
            raise WriteError(f"could not create server data directory, {str(e)}")

        with open(f"{path}/install.sh", "w", newline="\n") as f:
            f.write("cd /server\n" + "\n".join(image["installation"]["script"]))

        if spinner:
            spinner.info(
                "Pulling Docker image and creating installation container, do not exit"
            )
            spinner.start()

        self._docker_client.containers.run(
            image["installation"]["docker_image"],
            f"{image['installation']['shell']} /server/install.sh",
            volumes={path: {"bind": "/server", "mode": "rw"}},
            name=f"wilfred_{server.id}",
            environment=ContainerVariables(server, image, install=True).get_env_vars(),
            remove=True,
            detach=True,
        )

        if skip_wait and spinner:
            spinner.info(
                "Installation will continue in background, use `wilfred servers` to see if process has finished."
            )
            spinner.start()

        if not skip_wait:
            if spinner:
                spinner.info(
                    "You can safely press CTRL+C, the installation will continue in the background."
                )
                spinner.info(
                    "Run `wilfred servers` too see when the status changes from `installing` to `stopped`."
                )
                spinner.info(
                    f"You can also follow the installation log using `wilfred console {server.name}`"
                )
                spinner.start()
            while self._container_alive(server):
                sleep(1)

    def kill(self, server):
        """
        Kills server container

        Args:
            server (wilfred.database.Server): Server database object

        Raises:
            :py:class:`ServerNotRunning`
                If server is not running
        """

        try:
            container = self._docker_client.containers.get(f"wilfred_{server.id}")
        except docker.errors.NotFound:
            raise ServerNotRunning(f"server {server.id} is not running")

        container.kill()

    def rename(self, server, name):
        """
        Renames server and moves server folder

        Args:
            server (wilfred.database.Server): Server database object
            name (str): New name of the server

        Raises:
            :py:class:`WilfredException`
                If server is running
            :py:class:`WriteError`
                If not able to move folder
        """

        if self._container_alive(server):
            raise WilfredException("You cannot rename the server while it is running")

        try:
            rename(
                f"{self._configuration['data_path']}/{server.name}_{server.id}",
                f"{self._configuration['data_path']}/{name}_{server.id}",
            )
        except Exception as e:
            raise WriteError(f"could not rename folder, {str(e)}")

        server.name = name
        session.commit()

    def _console_input_callback(self, payload, server):
        self.command(server, payload)

    def command(self, server, command):
        """
        Sends command to server console

        Args:
            server (wilfred.database.Server): Server database object
            command (str): The command to send to the stdin of the server

        Raises:
            :py:class:`ServerNotRunning`
                If server is not running
        """

        _cmd = f"{command}\n".encode("utf-8")

        try:
            container = self._docker_client.containers.get(f"wilfred_{server.id}")
        except docker.errors.NotFound:
            raise ServerNotRunning(f"server {server.id} is not running")

        s = container.attach_socket(params={"stdin": 1, "stream": 1})
        s.send(_cmd) if platform.startswith("win") else s._sock.send(_cmd)
        s.close()

    def _running_docker_sync(self):
        for server in session.query(Server).all():
            try:
                self._docker_client.containers.get(f"wilfred_{server.id}")
            except docker.errors.NotFound:
                self.set_status(server, "stopped")

    def _parse_startup_command(self, cmd, server, image):
        return ContainerVariables(server, image).parse_startup_command(
            cmd.replace("{{SERVER_MEMORY}}", str(server.memory)).replace(
                "{{SERVER_PORT}}", str(server.port)
            )
        )

    def _container_alive(self, server):
        try:
            self._docker_client.containers.get(f"wilfred_{server.id}")
        except docker.errors.NotFound:
            return False

        return True

    def _start(self, server):
        path = f"{self._configuration['data_path']}/{server.name}_{server.id}"
        image = self._images.get_image(server.image_uid)

        try:
            remove_file(f"{path}/install.sh")
        except Exception:
            pass

        self._docker_client.containers.run(
            image["docker_image"],
            self._parse_startup_command(server.custom_startup, server, image)
            if server.custom_startup is not None
            else f"{self._parse_startup_command(image['command'], server, image)}",
            volumes={path: {"bind": "/server", "mode": "rw"}},
            name=f"wilfred_{server.id}",
            remove=True,
            ports={f"{server.port}/tcp": server.port},
            detach=True,
            working_dir="/server",
            mem_limit=f"{server.memory}m",
            oom_kill_disable=True,
            stdin_open=True,
            environment=ContainerVariables(server, image).get_env_vars(),
            user=image["user"] if image["user"] else "root",
        )

    def _stop(self, server):
        image = self._images.get_image(server.image_uid)

        try:
            container = self._docker_client.containers.get(f"wilfred_{server.id}")
        except docker.errors.NotFound:
            return

        if not image["stop_command"]:
            container.stop()

            return

        self.command(server, image["stop_command"])

        stopped = False

        while not stopped:
            try:
                self._docker_client.containers.get(f"wilfred_{server.id}")
            except docker.errors.NotFound:
                stopped = True



# -*- coding: utf-8 -*-

####################################################################
#                                                                  #
# Wilfred                                                          #
# Copyright (C) 2020, Vilhelm Prytz, <vilhelm@prytznet.se>, et al. #
#                                                                  #
# Licensed under the terms of the MIT license, see LICENSE.        #
# https://github.com/wilfred-dev/wilfred                           #
#                                                                  #
####################################################################

import click
import codecs
import locale
import os
import sys

from halo import Halo
from pathlib import Path
from sqlalchemy.exc import IntegrityError
from sqlalchemy import inspect
from time import sleep
from tabulate import tabulate
from shutil import get_terminal_size

from wilfred.docker_conn import docker_client
from wilfred.version import version, commit_hash, commit_date
from wilfred.api.config_parser import Config, NoConfiguration
from wilfred.database import session, database_path, Server, EnvironmentVariable
from wilfred.api.servers import Servers
from wilfred.api.images import Images, ImageAPIMismatch, ImagesOutdated
from wilfred.message_handler import warning, error, ui_exception
from wilfred.core import is_integer, random_string, check_for_new_releases
from wilfred.migrate import Migrate
from wilfred.api.server_config import ServerConfig
from wilfred.decorators import configuration_present


config = Config()

try:
    config.read()
except NoConfiguration:
    warning("Wilfred is not yet configured. Run `wilfred setup` to configure Wilfred.")
except Exception as e:
    ui_exception(e)


images = Images()

if not os.environ.get("WILFRED_SKIP_DOCKER", False):
    try:
        servers = Servers(docker_client(), config.configuration, images)
    except Exception as e:
        ui_exception(e)

if not images.check_if_present():
    with Halo(
        text="Downloading default images", color="yellow", spinner="dots"
    ) as spinner:
        images.download()
        spinner.succeed("Images downloaded")

try:
    images.read_images()
except ImageAPIMismatch:
    with Halo(
        text="Downloading default images", color="yellow", spinner="dots"
    ) as spinner:
        images.download()
        spinner.succeed("Images downloaded")
    try:
        images.read_images()
    except Exception as e:
        ui_exception(e)
except ImagesOutdated:
    with Halo(
        text="Images outdated, refreshing default images",
        color="yellow",
        spinner="dots",
    ) as spinner:
        images.download()
        spinner.succeed("Images refreshed")
    try:
        images.read_images()
    except Exception as e:
        ui_exception(e)
except Exception as e:
    ui_exception(e)

# check
Migrate()

ENABLE_EMOJIS = False if sys.platform.startswith("win") else True


def print_version(ctx, param, value):
    """
    print version and exit
    """

    if not value or ctx.resilient_parsing:
        return

    _commit_hash = commit_hash[0:7]
    _snap = (
        f" via snap (revision {os.environ['SNAP_REVISION']})"
        if "SNAP" in os.environ and "SNAP_REVISION" in os.environ
        else ""
    )

    check_for_new_releases(enable_emojis=ENABLE_EMOJIS)
    if str(version) == "0.0.0.dev0":
        click.echo(
            "".join(
                (
                    f"{'✨ ' if ENABLE_EMOJIS else ''}wilfred version ",
                    f"{_commit_hash}/edge (development build) built {commit_date}{_snap}",
                )
            )
        )
    else:
        click.echo(
            "".join(
                (
                    f"{'✨ ' if ENABLE_EMOJIS else ''}wilfred version ",
                    f"v{version}/stable (commit {_commit_hash}) built {commit_date}{_snap}",
                )
            )
        )

    ctx.exit()


def print_path(ctx, param, value):
    """
    print config/data paths
    """

    if not value or ctx.resilient_parsing:
        return

    click.echo(f"Configuration file: {click.format_filename(config.config_path)}")
    click.echo(f"Image config file: {click.format_filename(images.image_dir)}")
    click.echo(f"Database file: {click.format_filename(database_path)}")

    if config.configuration:
        _path = f"{click.format_filename(config.configuration['data_path'])}"

        if sys.platform.startswith("win"):
            _path = _path.replace("/", "\\")

        click.echo(f"Server data: {_path}")

    ctx.exit()


def pretty_list(data, tablefmt):
    for server in data:
        server.update(
            (
                k,
                str(v)
                .replace("running", click.style("running", fg="green"))
                .replace("stopped", click.style("stopped", fg="red"))
                .replace("installing", click.style("installing", fg="yellow")),
            )
            for k, v in server.items()
        )

    headers = {
        "id": click.style("ID", bold=True),
        "name": click.style("Name", bold=True),
        "image_uid": click.style("Image UID", bold=True),
        "memory": click.style("RAM", bold=True),
        "port": click.style("Port", bold=True),
        "status": click.style("Status", bold=True),
        "custom_startup": click.style("Custom startup", bold=True),
    }

    return tabulate(
        data,
        headers=headers,
        tablefmt=tablefmt,
    )


def main():
    # snap packages raise some weird ASCII codec errors, so we just force C.UTF-8
    if (
        codecs.lookup(locale.getpreferredencoding()).name == "ascii"
        and os.name == "posix"
    ):
        os.environ["LC_ALL"] = "C.UTF-8"
        os.environ["LANG"] = "C.UTF-8"

    cli()


@click.group()
@click.option(
    "--version",
    is_flag=True,
    callback=print_version,
    expose_value=False,
    is_eager=True,
    help="Print version and exit",
)
@click.option(
    "--path",
    is_flag=True,
    callback=print_path,
    expose_value=False,
    is_eager=True,
    help="Print paths for configurations and server data",
)
def cli():
    """
    Wilfred - A CLI for managing game servers using Docker.

    Website - https://wilfredproject.org

    Official documentation - https://docs.wilfredproject.org

    Source code - https://github.com/wilfred-dev/wilfred

    Discord server for support - https://wilfredproject.org/discord
    """

    pass


@cli.command()
def setup():
    """Setup wilfred, create configuration."""

    if config.configuration:
        warning("A configuration file for Wilfred already exists.")
        click.confirm("Are you sure you wan't to continue?", abort=True)

    data_path = click.prompt(
        "Path for storing server data",
        default=f"{str(Path.home())}/wilfred-data/servers",
    )

    config.write(data_path)


@cli.command("servers")
def servers_list():
    """List all existing servers."""

    # run sync to refresh server state
    servers.sync()

    data = servers.all()

    click.echo(
        pretty_list(
            data,
            tablefmt="plain" if get_terminal_size((80, 20))[0] < 96 else "fancy_grid",
        )
    )


@cli.command("images")
@click.option("--refresh", help="Download the default images from GitHub", is_flag=True)
@click.option(
    "--repo",
    help="Specify repo to fetch images from during image refresh",
    default="wilfred-dev/images",
    show_default=True,
)
@click.option(
    "--branch",
    help="Specify branch to fetch images from during image refresh",
    default="master",
    show_default=True,
)
def list_images(refresh, repo, branch):
    """List images available on file."""

    if refresh:
        with Halo(
            text=f"Refreshing images [{repo}/{branch}]", color="yellow", spinner="dots"
        ) as spinner:
            try:
                images.download(repo=repo, branch=branch)
                images.read_images()
            except Exception as e:
                spinner.fail()
                ui_exception(e)

            spinner.succeed(f"Images refreshed [{repo}/{branch}]")

    click.echo(
        tabulate(
            images.data_strip_non_ui(),
            headers={
                "uid": click.style("UID", bold=True),
                "name": click.style("Image Name", bold=True),
                "author": click.style("Author", bold=True),
                "default_image": click.style("Default Image", bold=True),
            },
            tablefmt="fancy_grid",
        )
    )

    updated_at = click.style(
        images.image_fetch_date.strftime("%Y-%m-%d %H:%M:%S"), bold=True
    )
    will_refresh_in = click.style(
        str(images.image_time_to_refresh).split(".")[0], bold=True
    )

    click.echo(
        f"Default images last updated at {updated_at}, will refresh in {will_refresh_in}"
    )


@cli.command()
@click.option(
    "--console",
    help="Attach to server console immediately after creation.",
    is_flag=True,
)
@click.option(
    "--detach",
    help="Immediately detach during install.",
    is_flag=True,
)
@click.pass_context
@configuration_present
def create(ctx, console, detach):
    """Create a new server."""

    name = click.prompt("Server Name").lower()

    if " " in name:
        error("space not allowed in name", exit_code=1)

    click.secho("Available Images", bold=True)
    click.echo(
        tabulate(
            images.data_strip_non_ui(),
            headers={
                "uid": click.style("UID", bold=True),
                "name": click.style("Image Name", bold=True),
                "author": click.style("Author", bold=True),
                "default_image": click.style("Default Image", bold=True),
            },
            tablefmt="fancy_grid",
        )
    )

    image_uid = click.prompt("Image UID", default="minecraft-vanilla")

    if " " in image_uid:
        error("space not allowed in image_uid", exit_code=1)

    if not images.get_image(image_uid):
        error("image does not exist", exit_code=1)

    port = click.prompt("Port", default=25565)
    memory = click.prompt("Memory", default=1024)

    # create
    server = Server(id=random_string())

    try:
        server.name = name
    except ValueError as e:
        error(str(e), exit_code=1)

    server.image_uid = image_uid
    server.memory = memory
    server.port = port
    server.custom_startup = None
    server.status = "installing"

    session.add(server)

    try:
        session.commit()
    except IntegrityError as e:
        error(f"unable to create server {click.style(str(e), bold=True)}", exit_code=1)

    click.secho("Environment Variables", bold=True)

    # environment variables available for the container
    for v in images.get_image(image_uid)["variables"]:
        if not v["hidden"]:
            value = click.prompt(
                v["prompt"], default=v["default"] if v["default"] is not True else None
            )

        if v["hidden"]:
            value = v["default"]

        variable = EnvironmentVariable(
            server_id=server.id, variable=v["variable"], value=value
        )
        session.add(variable)

    try:
        session.commit()
    except IntegrityError as e:
        error(
            f"unable to create variables {click.style(str(e), bold=True)}", exit_code=1
        )

    # custom startup command
    if click.confirm("Would you like to set a custom startup command (optional)?"):
        custom_startup = click.prompt(
            "Custom startup command", default=images.get_image(image_uid)["command"]
        )

        server.custom_startup = custom_startup
        try:
            session.commit()
        except IntegrityError as e:
            error(
                f"unable to set startup command {click.style(str(e), bold=True)}",
                exit_code=1,
            )

    with Halo(text="Creating server", color="yellow", spinner="dots") as spinner:
        try:
            servers.install(
                server, skip_wait=True if detach else False, spinner=spinner
            )
        except Exception as e:
            spinner.fail()
            ui_exception(e)
        spinner.succeed("Server created")

    if console:
        ctx.invoke(start, name=name)
        ctx.invoke(server_console, name=name)


@cli.command("sync")
@configuration_present
def sync_cmd():
    """
    Sync all servers on file with Docker (start/stop/kill).
    """

    with Halo(text="Docker sync", color="yellow", spinner="dots") as spinner:
        try:
            servers.sync()
        except Exception as e:
            spinner.fail()
            ui_exception(e)
        spinner.succeed("Servers synced")


@cli.command(short_help="Start server")
@click.argument("name")
@click.option(
    "--console", help="Attach to server console immediately after start.", is_flag=True
)
@click.pass_context
@configuration_present
def start(ctx, name, console):
    """
    Start server

    NAME is the name of the server
    """

    try:
        servers.sync()
    except Exception as e:
        ui_exception(e)

    with Halo(text="Starting server", color="yellow", spinner="dots") as spinner:
        server = session.query(Server).filter_by(name=name.lower()).first()

        if not server:
            spinner.fail("Server does not exit")
            sys.exit(1)

        if server.status == "installing":
            spinner.fail("Server is installing, start blocked.")
            sys.exit(1)

        image = images.get_image(server.image_uid)

        if not image:
            error("Image UID does not exit", exit_code=1)

        try:
            ServerConfig(
                config.configuration, servers, server, image
            ).write_environment_variables()
        except Exception as e:
            ui_exception(e)

        try:
            servers.set_status(server, "running")
            servers.sync()
        except Exception as e:
            spinner.fail()
            ui_exception(e)

        spinner.succeed("Server started")

        if console:
            ctx.invoke(server_console, name=name)


@cli.command(short_help="Forcefully kill running server")
@click.argument("name")
@click.option("-f", "--force", is_flag=True, help="Force action without confirmation")
@configuration_present
def kill(name, force):
    """
    Forcefully kill running server

    NAME is the name of the server
    """

    if force or click.confirm(
        "Are you sure you want to do this? This will kill the running container without saving data."
    ):
        with Halo(text="Killing server", color="yellow", spinner="dots") as spinner:
            server = session.query(Server).filter_by(name=name.lower()).first()

            if not server:
                spinner.fail("Server does not exit")
                sys.exit(1)

            try:
                servers.kill(server)
                servers.set_status(server, "stopped")
                servers.sync()
            except Exception as e:
                spinner.fail()
                ui_exception(e)

            spinner.succeed("Server killed")


@cli.command(short_help="Stop server gracefully")
@click.argument("name")
@configuration_present
def stop(name):
    """
    Stop server gracefully.

    NAME is the name of the server
    """

    try:
        servers.sync()
    except Exception as e:
        ui_exception(e)

    with Halo(text="Stopping server", color="yellow", spinner="dots") as spinner:
        server = session.query(Server).filter_by(name=name.lower()).first()

        if not server:
            spinner.fail("Server does not exit")
            sys.exit(1)

        if server.status == "installing":
            spinner.fail(
                " ".join(
                    (
                        "Server is installing, you cannot gracefully stop it.",
                        "Use `wilfred kill` if the installation process has hanged.",
                    )
                )
            )
            sys.exit(1)

        try:
            servers.set_status(server, "stopped")
            servers.sync()
        except Exception as e:
            spinner.fail()
            ui_exception(e)

        spinner.succeed("Server stopped")


@cli.command(short_help="Restart server")
@click.argument("name")
@click.option(
    "--console", help="Attach to server console immediately after start.", is_flag=True
)
@click.pass_context
@configuration_present
def restart(ctx, name, console):
    """
    Restart server

    NAME is the name of the server
    """

    ctx.invoke(stop, name=name)
    ctx.invoke(start, name=name)

    if console:
        ctx.invoke(server_console, name=name)


@cli.command()
@click.argument("name")
@click.option("-f", "--force", is_flag=True, help="Force action without confirmation")
@configuration_present
def delete(name, force):
    """
    Delete existing server.

    NAME is the name of the server
    """

    if force or click.confirm(
        "Are you sure you want to do this? All data will be permanently deleted."
    ):
        with Halo(text="Deleting server", color="yellow", spinner="dots") as spinner:
            server = session.query(Server).filter_by(name=name.lower()).first()

            if not server:
                spinner.fail("Server does not exit")
                sys.exit(1)

            try:
                servers.remove(server)
                spinner.succeed("Server removed")
            except Exception as e:
                spinner.fail()
                ui_exception(e)


@cli.command("command", short_help="Send command to STDIN of server")
@click.argument("name")
@click.argument("command")
@configuration_present
def run_command(name, command):
    """
    Send command to STDIN of server

    \b
    NAME is the name of the server
    COMMAND is the command to send, can be put in \" for commands with whitespaces
    """

    server = session.query(Server).filter_by(name=name.lower()).first()

    if not server:
        error("Server does not exit", exit_code=1)

    try:
        servers.command(server, command)
    except Exception as e:
        ui_exception(e)


@cli.command(
    "console",
    short_help="\n\n".join(
        (
            " ".join(
                (
                    "Attach current terminal to the console of a specific server (STDIN of running process).",
                    "This allows you to view the current log and send commands.",
                )
            ),
            "Use CTRL+C to exit the console, the server will continue to run in the background.",
        )
    ),
)
@click.argument("name")
@configuration_present
def server_console(name):
    """
    Attach to server console, view log and run commands

    NAME is the name of the server
    """

    server = session.query(Server).filter_by(name=name.lower()).first()

    if not server:
        error("Server does not exit", exit_code=1)

    click.secho(
        " ".join(
            (
                f"Viewing server console of {name} (id {server.id})",
                f"{'- input disabled, installing' if server.status == 'installing' else ''}",
            )
        ),
        bold=True,
    )

    try:
        servers.console(
            server, disable_user_input=True if server.status == "installing" else False
        )
    except Exception as e:
        ui_exception(e)


@cli.command(
    short_help=" ".join(
        (
            "Edit existing server (name, memory, port, environment variables and the custom startup command).",
            "Restart server after update for changes to take effect.",
        )
    )
)
@click.argument("name")
@configuration_present
def edit(name):
    """
    Edit server (name, memory, port, environment variables)

    NAME is the name of the server
    """

    server = session.query(Server).filter_by(name=name.lower()).first()

    if not server:
        error("Server does not exist", exit_code=1)

    click.echo(
        pretty_list(
            [
                {
                    c.key: getattr(server, c.key)
                    for c in inspect(server).mapper.column_attrs
                }
            ],
            tablefmt="plain" if get_terminal_size((80, 20))[0] < 96 else "fancy_grid",
        )
    )
    click.echo("Leave values empty to use existing value")

    name = click.prompt("Name", default=server.name).lower()

    if " " in name:
        error("space not allowed in name", exit_code=1)

    port = click.prompt("Port", default=server.port)
    memory = click.prompt("Memory", default=server.memory)

    if not is_integer(port) or not is_integer(memory):
        error("port/memory must be integer", exit_code=1)

    click.secho("Environment Variables", bold=True)

    for v in images.get_image(server.image_uid)["variables"]:
        if v["install_only"]:
            continue

        current_variable = (
            session.query(EnvironmentVariable)
            .filter_by(server_id=server.id)
            .filter_by(variable=v["variable"])
            .first()
        )

        if not current_variable:
            continue

        if not v["hidden"]:
            value = click.prompt(v["prompt"], default=current_variable.value)

        if v["hidden"]:
            value = v["default"]

        current_variable.value = value

        try:
            session.commit()
        except IntegrityError as e:
            error(
                f"unable to edit variables {click.style(str(e), bold=True)}",
                exit_code=1,
            )

    server.custom_startup = (
        None if server.custom_startup == "None" else server.custom_startup
    )
    custom_startup = None

    if server.custom_startup is not None:
        custom_startup = click.prompt(
            "Custom startup command (use 'None' to reset to default)",
            default=server.custom_startup,
        )

    if server.custom_startup is None:
        if click.confirm("Would you like to set a custom startup command?"):
            custom_startup = click.prompt(
                "Custom startup command (use 'None' to reset to default)",
                default=images.get_image(server.image_uid)["command"],
            )

            custom_startup = None if custom_startup == "None" else custom_startup

    if name != server.name:
        try:
            servers.rename(server, name)
        except Exception as e:
            ui_exception(e)

    server.port = port
    server.memory = memory

    if custom_startup:
        server.custom_startup = custom_startup

    try:
        session.commit()
    except IntegrityError as e:
        error(f"unable to edit server {click.style(str(e), bold=True)}", exit_code=1)

    click.echo(
        "✅ Server information updated, restart server for changes to take effect"
    )


@cli.command(
    short_help=" ".join(("Show server statistics, CPU load, memory load etc.",))
)
def top():
    while True:
        # retrieve data (this can take a split moment)
        data = servers.all(cpu_load=True, memory_usage=True)

        for server in data:
            server.update(
                (
                    k,
                    str(v)
                    .replace("running", click.style("running", fg="green"))
                    .replace("stopped", click.style("stopped", fg="red"))
                    .replace("installing", click.style("installing", fg="yellow")),
                )
                for k, v in server.items()
            )

        # clear the screen
        click.clear()

        headers = {
            "id": click.style("ID", bold=True),
            "name": click.style("Name", bold=True),
            "image_uid": click.style("Image UID", bold=True),
            "memory": click.style("RAM", bold=True),
            "port": click.style("Port", bold=True),
            "status": click.style("Status", bold=True),
            "custom_startup": click.style("Custom startup", bold=True),
            "cpu_load": click.style("CPU", bold=True),
            "memory_usage": click.style("RAM usage", bold=True),
        }

        # display table
        click.echo(
            tabulate(
                data,
                headers=headers,
                tablefmt="plain",
            )
        )

        # cooldown before repeating
        sleep(1)


@cli.command(
    "config",
    short_help="".join(
        (
            "View and edit configuration variables (e.g. the settings in `server.properties` for Minecraft).",
        )
    ),
)
@click.argument("name")
@click.argument("variable", required=False)
@click.argument("value", required=False)
def config_command(name, variable, value):
    """
    Manage server configuration (for supported filetypes)

    \b
    NAME is the name of the server
    VARIABLE is the name of an available setting
    VALUE is the new value for the variable setting
    """

    def _get():
        return ServerConfig(config.configuration, servers, server, image)

    def _print_all_values(variable, config_list):
        for var in config_list:
            click.echo(
                f"{click.style(var['_wilfred_config_filename'], bold=True)} {variable}: '{var[variable]}'"
            )

    def _get_variable_occurrences(variable, raw):
        _variable_occurrences = []
        for x in raw:
            if variable in x:
                _variable_occurrences.append(x)

        return _variable_occurrences

    server = session.query(Server).filter_by(name=name.lower()).first()

    if not server:
        error("Server does not exist", exit_code=1)

    image = images.get_image(server.image_uid)

    if not image:
        error("Image UID does not exit", exit_code=1)

    server_conf = _get()

    _variable_occurrences = _get_variable_occurrences(variable, server_conf.raw)

    if variable and len(_variable_occurrences) == 0:
        error("variable does not exist", exit_code=1)

    if variable and len(_variable_occurrences) > 1:
        click.echo("This variable exists in multiple configuration files.")

    if variable and not value:
        _print_all_values(variable, _variable_occurrences)
        exit(0)

    if variable and value:
        user_selection = None
        if len(_variable_occurrences) > 1:
            for i in range(len(_variable_occurrences)):
                click.echo(
                    f"[{i}] - {_variable_occurrences[i]['_wilfred_config_filename']}"
                )

            user_selection = click.prompt(
                "In which file would like to modify this setting?", default=0
            )

            if (
                int(user_selection) >= len(_variable_occurrences)
                or int(user_selection) < 0
            ):
                error(
                    "integer is not valid, please pick one from the list", exit_code=1
                )

        filename = _variable_occurrences[user_selection if user_selection else 0][
            "_wilfred_config_filename"
        ]

        try:
            server_conf.edit(filename, variable, value)
            server_conf = _get()
            _print_all_values(
                variable, _get_variable_occurrences(variable, server_conf.raw)
            )
        except Exception as e:
            ui_exception(e)

        exit(0)

    click.echo(server_conf.pretty())


if __name__ == "__main__":
    main()



####################################################################
#                                                                  #
# Wilfred                                                          #
# Copyright (C) 2020, Vilhelm Prytz, <vilhelm@prytznet.se>, et al. #
#                                                                  #
# Licensed under the terms of the MIT license, see LICENSE.        #
# https://github.com/wilfred-dev/wilfred                           #
#                                                                  #
####################################################################

import click
import sys


def _message(prefix, msg):
    click.echo(f"{prefix} {msg}")


def ui_exception(e):
    _exception_name = click.style(f"{type(e).__name__}", bold=True)
    error(f"{_exception_name} {str(e)}", exit_code=1)


def error(message, exit_code=None):
    _message(prefix=click.style("💥 Error", fg="red"), msg=message)

    if exit_code is not None:
        sys.exit(exit_code)


def warning(message, exit_code=None):
    _message(prefix=click.style("⚠️  Warning ", fg="yellow"), msg=message)

    if exit_code is not None:
        sys.exit(exit_code)


def info(message, exit_code=None):
    _message(prefix=click.style("🔵 Info", fg="blue"), msg=message)

    if exit_code is not None:
        sys.exit(exit_code)
